from PIL import Image
from typing import NamedTuple, Optional
from ..config import ConfigurationMyOdt
import os
import zipfile
import shutil
import logging
import datetime
import time


# logger = ConfigurationMyOdt.logger

class ODTMeta(NamedTuple):
    """Tuple-type container class for storing and quickly handling information """
    name: str  # Имя документа.
    document: bytes  # Документ в виде массива байтов.
    text_replace: list  # Список словарей ключь-значения для замены.
    extend_set_text: Optional[list]  # Список словарей ключь-значения для замены в опциях {%extend_set()%}.
    diagram: Optional[list]  # Список содержащий данные для работы с диаграмой.
    images: Optional[list]  # Список содержащий словари словари с опциями для вставки изображений
    options: Optional[list]  # Список опций сгенерированных из входных данных
    if_options: Optional[dict]


class Handler:
    """This class contains auxiliary methods"""

    @staticmethod
    def check_slash(path: str) -> str:
        """
        This method checks for a closing slash at the end of the path. Example: a = '/home/user/folder'
        a = Handler.chek_slash(a) a -> /home/user/folder/
        """
        if not path[-1] == '/':
            path = path + '/'
        return path

    @staticmethod
    def check_filename(file_name: str) -> str:
        """
        This method checks the file name for an extension and returns the name without an extension. Example:
        a = 'Filename.odt'
        a = Handler.check_filename(a)
        a -> 'Filename'
        """
        file_name = file_name.split('.')[0]
        return file_name

    @staticmethod
    def check_format(format_file: str) -> str:
        """
        This method checks the parameter containing the file extension. Example: a = 'ODT'
        a = Handler.check_format(a)
        a -> '.odt'
        """
        format_file = format_file.lower()
        if not format_file[0] == '.':
            format_file = '.' + format_file
        return format_file

    @staticmethod
    def is_exist(path: str) -> bool:
        """This method checks the existence of a path and returns True or False"""
        if os.path.exists(path):
            return True
        else:
            return False

    @staticmethod
    def make_folder(path: str) -> None:
        """Use this method to create a file or folder."""
        try:
            os.makedirs(path)
        except Exception:
            # logger.exception('lib > Handler > make_folder > The {} directory could not be created'.format(path))
            raise

    @staticmethod
    def check_and_make(path: str) -> None:
        """This method checks for the specified file or directory and creates it if it is not present."""
        if Handler.is_exist(path):
            logging.debug('lib > Handler > check_and_make > Directory {} already exists'.format(path))
        else:
            logging.debug('lib > Handler > check_and_make > Directory {} does not exist'.format(path))
            Handler.make_folder(path)

    @staticmethod
    def copyfile(copy_file: str, out_folder: str, out_name: str, out_format: str) -> None:
        """
        The method checks if there is a file and copies it to the specified folder
        with the specified name and extension.
        """
        if Handler.is_exist(copy_file):
            try:
                Handler.check_and_make(out_folder)
                shutil.copyfile(copy_file, Handler.check_slash(out_folder) + Handler.check_filename(
                    out_name) + Handler.check_format(out_format))
            except Exception:
                # logger.exception('lib > Handler > copyfile > File {} copying crash'.format(copy_file))
                raise
        # else:
        #     logging.error('lib > Handler > copyfile > {} - no such file or directory'.format(copy_file))

    @staticmethod
    def movefile(copy_file: str, out_folder: str, out_name: str, out_format: str) -> None:
        """
        The method checks if there is a file and movies it to the specified folder
        with the specified name and extension.
        """
        if Handler.is_exist(copy_file):
            try:
                Handler.check_and_make(out_folder)
                shutil.copyfile(copy_file,
                                Handler.check_slash(out_folder) + Handler.check_filename(
                                    out_name) + Handler.check_format(out_format))
                os.remove(copy_file)
            except Exception:
                # logger.exception('lib > Handler > movefile > File {} movings crash'.format(copy_file))
                raise
        else:
            pass
            # logging.error('lib > Handler > movefile > {} - no such file or directory'.format(copy_file))

    @staticmethod
    def removefile(file: str) -> None:
        """This method deletes the specified file or directory."""
        if Handler.is_exist(file):
            try:
                if '.' in file.split('/')[-1]:
                    os.remove(file)
                else:
                    shutil.rmtree(file)
            except Exception:
                # logger.exception('lib > Handler > removefile > File {} remove crash'.format(file))
                raise
        # else:
            # logging.error('lib > Handler > removefile > {} - no such file or directory'.format(file))

    @staticmethod
    def renamefile(file_old: str, file_new: str) -> None:
        """This method renames the specified file or directory."""
        if Handler.is_exist(file_old):
            try:
                os.rename(file_old, file_new)
            except Exception:
                # logger.exception('lib > Handler > renamefile > File {} rename to {} crash'.format(file_old, file_new))
                raise
        # else:
        #     logging.error('lib > Handler > renamefile > {} - no such file or directory'.format(file_old))


class Archiver:
    """Class for Archive and Unarchive *.odt templates"""

    @staticmethod
    def extract_template(file_path: str, file_name: str, out_path: str) -> None:
        """
        Unpacks the zip archive
        Example: /home/user/Archive1.zip -> /home/out/{ARCHIVE NAME}
        file_path: Path to the folder containing the file. # /home/user/
        file_name: File name without file extension. # Archive1
        out_path: Path to the folder to unpack to, in a subdirectory equal to the archive name. # /home/out/
        """
        file_path = Handler.check_slash(file_path)
        file_name = Handler.check_filename(file_name)
        out_path = Handler.check_slash(out_path)
        full_path = file_path + file_name
        full_out_path = out_path + file_name
        try:
            fantasy_zip = zipfile.ZipFile(full_path + '.zip')
            fantasy_zip.extractall(full_out_path)
            fantasy_zip.close()
        except Exception:
            # logger.exception('lib > Archiver > execute_template > File: {} not extracting'.format(full_path))
            raise

    @staticmethod
    def pack_template(path_to_pack_folder: str, out_path: str, out_name: str) -> None:
        """
        To pack the files in a directory to the archive.
        path_to_pack_folder: The path to the directory with the backup files.
        out_path: The path where the archive will be saved.
        out_name: Archive name without extension.
        """
        path_to_pack_folder = Handler.check_slash(path_to_pack_folder)
        out_path = Handler.check_slash(out_path)
        out_name = Handler.check_filename(out_name)
        try:
            Handler.check_and_make(out_path)
            z = zipfile.ZipFile(out_path + out_name + '.zip', 'w')  # Создание нового архива
            for root, dirs, files in os.walk(path_to_pack_folder):  # Список всех файлов и папок в директории folder
                without_root = root.replace(path_to_pack_folder, "", 1)
                for file in files:
                    # Создание относительных путей и запись файлов в архив
                    z.write(os.path.join(root, file), arcname=os.path.join(without_root, file))
            z.close()
        except Exception:
            # logger.exception('lib > Archiver > pack_template > File {} zip not complete'.format(out_name))
            raise


class ImageResize:
    def __init__(self, path_image_master: str, path_image_submissive: str, path_image_out: str,
                 border_color: tuple = (255, 255, 255)):
        """Класс приводит размеры изоброжение ведомого под размеры изображения мастера с сохранением пропорций.
        В path_image_master = /you`r_path/image_name.jpg передается полный или относительный путь с указанием имени,
        и расширения изображения. Аналогично и для переменных path_image_submissive, path_image_out.
        """
        self.img_master = path_image_master
        self.img_submissive = path_image_submissive
        self.border_color = border_color
        self.out_image = path_image_out
        self.master_size = None
        self.submissive_size = None
        self.master_im = None
        self.submissive_im = None

    def _image_size(self) -> None:
        """Функция открывает изображение и является сетером атриботов класса содержащих размеры изображения"""
        if (Handler.is_exist(self.img_master)) and (Handler.is_exist(self.img_submissive)):
            self.master_im = Image.open(self.img_master)
            self.master_size = dict(width=float(self.master_im.size[0]), height=float(self.master_im.size[1]))
            self.submissive_im = Image.open(self.img_submissive)
            self.submissive_size = dict(width=float(self.submissive_im.size[0]), height=float(self.submissive_im.size[1]))
        else:
            # logger.error('lib > ImageResize > _image_size > Error')
            raise Exception('Один или несколько файлов не найдены. {} {}'.format(self.img_master, self.img_submissive))

    def _check_size(self) -> None:
        """
        Математическая логика приведения размера сторон ведомого изображения, к сторонам ведущего.
        В процентном соотношении, с сохранением пропорций. Без привышения какой либо из сторон ведомого изображения,
        длинны стороны ведущего.
        """
        a1 = self.master_size['width']
        b1 = self.master_size['height']
        a2 = self.submissive_size['width']
        b2 = self.submissive_size['height']
        if ((a1 < a2) and (b1 == b2)) or ((a1 > a2) and (b1 == b2)):
            k = a1 / a2
            self.submissive_size['width'] = a2 * k
            self.submissive_size['height'] = b2 * k
        elif ((a1 == a2) and (b1 < b2)) or ((a1 == a2) and (b1 > b2)):
            k = b1 / b2
            self.submissive_size['width'] = a2 * k
            self.submissive_size['height'] = b2 * k
        elif ((a1 < a2) and (b1 < b2)) or ((a1 > a2) and (b1 > b2)):
            k1 = a1 / a2
            k2 = b1 / b2
            if k1 < k2:
                self.submissive_size['width'] = a2 * k1
                self.submissive_size['height'] = b2 * k1
            elif k2 < k1:
                self.submissive_size['width'] = a2 * k2
                self.submissive_size['height'] = b2 * k2
            elif k1 == k2:
                self.submissive_size['width'] = a2 * k1
                self.submissive_size['height'] = b2 * k1

    def apply_resize(self) -> None:
        """
        Основной, публичный метод, выполняющий расчеты и сохранение приведенного,
        к необходиммым параметрам, изображения.
        """
        try:
            self._image_size()
            self._check_size()
            master_size = (int(self.master_size['width']), int(self.master_size['height']))
            submissive_size = (int(self.submissive_size['width']), int(self.submissive_size['height']))
            new_im = Image.new('RGB', master_size, self.border_color)  # luckily, this is already black!
            self.submissive_im = self.submissive_im.resize((int(self.submissive_size['width']),
                                                            int(self.submissive_size['height'])))
            new_im.paste(self.submissive_im, (int((master_size[0]-submissive_size[0])/2),
                                              int((master_size[1]-submissive_size[1])/2)))
            # new_im.show()
            new_im.save(self.out_image)
        except Exception:
            # logger.error('lib > ImageResize > apply_resize > Error')
            raise


def runtime_hour(func):
    """A decorator that displays the execution time in the format hours, minutes, seconds, and microseconds."""
    def start_runtime_hour(*args, **kwargs):
        now = datetime.datetime.now()
        hour_start = now.hour
        min_start = now.minute
        sec_start = now.second
        msec_start = now.microsecond
        temp = func(*args, **kwargs)
        now = datetime.datetime.now()
        hour_stop = now.hour
        min_stop = now.minute
        sec_stop = now.second
        msec_stop = now.microsecond
        hour = hour_stop - hour_start
        min = min_stop - min_start
        sec = sec_stop - sec_start
        msec = msec_stop - msec_start
        print('Время выполнения составило: "ЧЧ:ММ:СС:МС" {}:{}:{}:{}'.format(hour, min, sec, msec))
        return temp
    return start_runtime_hour()


def runtime(func):
    """A decorator that displays the execution time in seconds with fractions."""
    def start_runtime(*args, **kwargs) -> str:
        now = time.time()
        temp = func(*args, **kwargs)
        print('Время выполнения составило: {}'.format(time.time()-now))
        return temp
    return start_runtime



import re
import copy
from ..config import ConfigurationMyOdt
from ..lib import ODTMeta
from ..lib import Handler

# logger = ConfigurationMyOdt.logger


class SettingExtendDiagram:
    def __init__(self, path: str, date: ODTMeta):
        self.path = Handler.check_slash(path)
        self.date = date

    @staticmethod
    def _parse_table(text_xml: str) -> list:
        """Функция разбора таблиц на составные части вида:
        table = [start_table,
                    ['<table:table-rows>', [
                        ['<table:table-row>', [<cell></cell>, <cell></cell>, <cell></cell>...], '</table:table-row>'],
                        ['<table:table-row>', [...], '</table:table-row>'],
                        ['<table:table-row>', [...], '</table:table-row>'],
                                           ],
                    '</table:table-rows>'],
                end_table]
        Где мы получем в table[1] список строк, где первая и последние строки являются тегами, а дальнейшие строки
        представляют из себя список ячеек <cell> каждой строки."""
        # Получаем из XML таблицу
        table = re.findall(r'<table.+?<\/table:table>', text_xml)[0]
        # Получаем начало таблицы
        table_start = re.findall(r'<table:table table:name.+?<\/table:table-header-rows>', table)[0]
        # Получаем строки, с данными для выделения конца таблицы.
        rows = re.findall(r'<table:table-rows>.+?<\/table:table-rows>', table)[0]
        # Получаем конец таблицы
        table_end = re.findall(r'<table:table-rows>.+?</table:table>', table)[0].replace(rows, '')
        # Получаем строки, с данными для заполнения списка
        row = re.findall(r'(<table:table-rows>.+?<\/table:table-rows>)', table)
        row = re.findall(r'(<table:table-row>.+?<\/table:table-row>)', row[0])
        table_new = [table_start, ['<table:table-rows>', [], '</table:table-rows>'], table_end]
        # Перебираем строки и выделяем ячейки.
        for i in range(len(row)):
            cell = re.findall(r'(<table:table-cell.+?<\/table:table-cell>)', row[i])
            temp = ['<table:table-row>', '</table:table-row>']
            temp.insert(1, cell)
            table_new[1][1].append(temp)
        return table_new

    @staticmethod
    def _check_data(table_option: dict, table: list) -> bool:
        """Функция проверяет колличество столбцов и строк во входных и выходных данных на соответсвие."""
        check = True
        if table_option['row'] == len(table[1][1]):
            for i in table[1][1]:
                if table_option['cell'] != len(i[1])-1:
                    check = False
        else:
            check = False
        return check

    @staticmethod
    def _add_row_and_cell(data_diagram: dict, table: list) -> list:
        """Метод добавляет строки в шаблон если занчений в параметрах больше чем в шаблоне"""
        if len(data_diagram['content']) > len(table[1][1]):
            multiplier = len(data_diagram['content']) - len(table[1][1])
            temp_row = table[1][1][-1]
            for i in range(multiplier):
                # Необходимо делать глубокое копирование строки, что бы различались id
                temp_row = copy.deepcopy(temp_row)
                table[1][1].append(temp_row)
        return table

    @staticmethod
    def _insert_data(data_in: str, cell_out: str) -> str:
        """
        Метод заменяет значения в переданной ячейке на новые
        data_in = 5555
        cell_out = <table:table-cell office:value-type="float" office:value="22222"><text:p>22222</text:p></table:table-cell>

        data_out = <table:table-cell office:value-type="float" office:value="555"><text:p>5555</text:p></table:table-cell>
        """
        value = re.findall(r'office:value="(.+?)"', cell_out)[0]
        data_out = cell_out.replace(value, data_in)
        return data_out

    @staticmethod
    def _insert_row_name(data_in: str, cell_in: str) -> str:
        """
        Метод заменяет значения в переданной ячейке на новые
        data_in = 5555
        cell_out = <table:table-cell office:value-type="float" office:value="22222"><text:p>22222</text:p></table:table-cell>

        data_out = <table:table-cell office:value-type="float" office:value="555"><text:p>5555</text:p></table:table-cell>
        """
        value = re.findall(r'<text:p>(.+?)</text:p>', cell_in)[0]
        data_out = cell_in.replace(value, data_in)
        return data_out

    @staticmethod
    def _check_first_cell(row: list) -> bool:
        """Метод проверяет на валидность первый элемент в списке."""
        list_index_str = []
        for cell_index in range(len(row)):
            value = re.findall(r'office:value-type="(.+?)"', row[cell_index])[0]
            if value == 'string':
                list_index_str.append(cell_index)
        if (len(list_index_str) == 1) and (list_index_str[0] == 0):
            return True
        else:
            return False

    @staticmethod
    def _loop_render(table: list, diagram: dict) -> list:
        """Метод перебирает строки, и ячейки в них, и вызывает фукнцию для замены значений на новые."""
        result_table = table
        for row in range(len(diagram['content'])):
            # Получаем строку  с входными данными [{'cell': '1', 'value': '1'}]
            row_in_data = diagram['content'][row]
            """
            Получаем строку с XML 
            ['<table:table-row>', 
            ['<table:table-cell office:value-type="string"><text:p>Строка 1</text:p></table:table-cell>', 
            '<table:table-cell office:value-type="float" office:value="9.1"><text:p>9.1</text:p></table:table-cell>'],
            '</table:table-row>']
            """
            # Меняем название строки.
            if row_in_data['row']:
                row_xml_table = result_table[1][1][row]
                result_table[1][1][row][1][0] = SettingExtendDiagram._insert_row_name(row_in_data['row'],
                                                                                      result_table[1][1][row][1][0])
            if SettingExtendDiagram._check_first_cell(row_xml_table[1]):
                for cell in range(len(row_xml_table[1])-1):
                    cell_in_data = row_in_data['cells'][cell]
                    cell_xml_data = row_xml_table[1][cell+1]
                    result_table[1][1][row][1][cell+1] = SettingExtendDiagram._insert_data(cell_in_data['value'],
                                                                                          cell_xml_data)
            else:
                # logger.error('odtapplysettings > setting_extend_diagram > _loop_render > Ошибка в типе первой ячейке '
                #              'входных данных. Или длинне строки. ' + row_xml_table)
                raise Exception('Ошибка в типе первой ячейке входных данных. Или длинне строки. ' + row_xml_table)
        return result_table

    @staticmethod
    def _chek_list_in_join(in_data: list) -> str:
        """Рекурсивный метод, объединяет структуру таблицы из листов в строку."""
        for i in range(len(in_data)):
            if type(in_data[i]) == list:
                temp = SettingExtendDiagram._chek_list_in_join(in_data[i])
                in_data[i] = temp
        out_data = ''.join(in_data)
        return out_data

    @staticmethod
    def _text_build(text_xml: str, table: str) -> str:
        """Метод для сборки текста. Заменяет текст таблицы в общем тексте документа"""
        # Получаем из XML таблицу
        table_search = re.findall(r'<table.+?<\/table:table>', text_xml)[0]
        text_xml = text_xml.replace(table_search, table)
        return text_xml

    def apply_setting(self) -> None:
        """Основной метод применения настроек"""
        # Перебираем диаграммы из входных данных
        for diagram in self.date.diagram:
            # Получаем имя диаграммы
            name = diagram['key_']
            # Берем порядковый номер ктороый присваивается в имени.
            name = name.split('_')[-1]
            # Создаем строку содержащую участок пути, например Object 1/content.xml
            name = 'Object ' + name + '/content.xml'
            # Проверяем есть ли файл по указанному пути
            if Handler.is_exist(self.path + name):
                # Если есть открываем
                with open(self.path + name, 'r') as file:
                    text = file.read()
            else:
                # Если нет то возвращаем ошибку
                # logger.error('odtapplysettings > setting_extend_diagram > _loop_render > The chart file in the ' +
                #              self.path + name + 'path does not exist! There may be an error in the chart name. ' +
                #              diagram['key_'])
                raise Exception('The chart file in the ' + self.path + name + 'path does not exist!'
                                ' There may be an error in the chart name. ' + diagram['key_'])
            # Вызываем функцию разбора таблицы на составные части в виде вложенных списков
            table = SettingExtendDiagram._parse_table(text)
            # Проверяем есть ли необходимость расширять строки, или же входные данные соответсвуют шаблону
            if not SettingExtendDiagram._check_data(self.date.options[0]['diagram'][diagram['key_']], table):
                table = SettingExtendDiagram._add_row_and_cell(diagram, table)
            table = SettingExtendDiagram._loop_render(table, diagram)
            table = SettingExtendDiagram._chek_list_in_join(table)
            text = SettingExtendDiagram._text_build(text, table)
            with open(self.path + name, 'w') as file:
                file.write(text)



import re
from ..config import ConfigurationMyOdt
from . import MyOdtCheckerAnchor
from ..lib import ODTMeta


# logger = ConfigurationMyOdt.logger


class SettingIfBlock:
    """The class uses the option to multiply rows in document tables."""

    def __init__(self, text_xml: str, date: ODTMeta):
        self.text = text_xml
        self.date = date
        self.if_block = []  # Объявляем пустую переменную типа list

    def _build_if_block_in_text(self):
        for key, value in self.date.if_options.items():
            if not value:
                regex_pattern = r"({%\s*if\s+" + re.escape(key) + r"\s*%}(.*?){%\s*endif\s*%})"
                regex = re.compile(regex_pattern, re.DOTALL)
                matches = regex.findall(text)
                self.text.replce(matches[0][0], '')
            else:
                regex_pattern = r"({%\s*if\s+" + re.escape(key) + r"\s*%}(.*?){%\s*endif\s*%})"
                regex = re.compile(regex_pattern, re.DOTALL)
                matches = regex.findall(text)
                self.text.replce(matches[0][0], matches[0][1])

    def apply_setting(self) -> str:
        """
        "ENG" The function applies all the methods of the class, and returns the result.

        "RUS" Функция применяет все  методы класса, и возвращает результат.
        """
        try:
            self._build_if_block_in_text()
            return self.text
        except Exception:
            # logger.exception(u'odt_engine > SettingExtendSet > apply_setting > Error')
            raise



import logging


class ConfigurationMyOdt:
    # logging.basicConfig(format=u'%(levelname)-8s [%(asctime)s] %(message)s', level=logging.DEBUG, filename=u'mylog.log')
    # logger = logging.getLogger()
    path_template_folder = './templates_odf/'

    # @classmethod
    # def set_loger(self, logger_in):
    #     self.logger = logger_in


import re


class MyOdtCheckerAnchor:
    """The class checks for inclusions in anchors in the document {{ANCHOR}} and {%options%}. Errors may occur during
    document generation, such as splitting the anchor into styles {%extend_set</text:span>(<text:span
    text:style-name= "T5">number=1, </text:span>extend_add=1)%}. In this case, the anchor will not be processed. The
    class removes anchor inclusions in the xml text. """

    @staticmethod
    def _check_anchor_list(anchor_list: list, text_xml: str) -> str:
        text_new = text_xml
        for i in range(len(anchor_list)):  # Перебираем список якорей
            # Делаем выборку включенных в якоря тегов XML {{PLACE_OF_<text:span text:style-name="T7">BIRTH</text:span>}}
            search_bag = re.findall(r'<.+?>', anchor_list[i])
            if len(search_bag) > 0:  # Если включения есть
                if search_bag[0][1:2] == '/':  # Проверяем первое включение, и если оно является закрывающим тегом XML
                    # Создаем переменную которая содержит регулярное выражения для поиска всего начиная с начала документа
                    # и до нашего якоря включительно.
                    search_var = '<.+?>' + anchor_list[i]
                    # Получаем кусок текста с начала документа и до нашего якоря
                    search_temp = re.findall(search_var, text_new)[0].replace(anchor_list[i], '')
                    # Получем список тегов с начала текста и до нашего якоря берем последний.
                    search_temp = re.findall(r'<.+?>', search_temp)[-1]
                    # В списке якорей в тексте добавляем открывающий XML тег который идет перед нашим якорем.
                    anchor_list[i] = search_temp + anchor_list[i]
                if search_bag[-1][
                   1:2] != '/':  # Проверяем последнее включение, и если оно является открывающим тегом XML
                    # Создаем переменную которая содержит регулярное выражения для поиска всего начиная с начала
                    # нашего якоря включительно и следующего за ним тега.
                    search_var = anchor_list[i] + '.+?>'
                    search_temp = re.findall(search_var, text_new)[0]  # Получаем кусок текста с описанного выше
                    # В списке якорей в тексте добавляем закрывающий XML тег который идет следом нашим якорем.
                    anchor_list[i] = search_temp
        for i in range(len(anchor_list)):  # Перебираем лист якорей
            search_bag = re.findall(r'<.+?>', anchor_list[i])  # Получаем лист багов
            if len(search_bag) > 0:  # Если лист багов существует
                for bag in search_bag:  # Бежим по нему
                    # Делаем временный лист якорей и убираем в нем баг
                    temp_token = anchor_list[i].replace(bag, '')
                    # Замещаем в тексте якорь с багами из листа на временный якорь с убранным багом
                    text_new = text_new.replace(anchor_list[i], temp_token)
                    # Убираем
                    anchor_list[i] = temp_token
        return text_new

    @staticmethod
    def check_anchor_bag(text_xml: str) -> str:
        new_text = text_xml
        search_anchor_list = re.findall(r'\{\{.+?\}\}', new_text)  # Получаем список якорей в тексте
        new_text = MyOdtCheckerAnchor._check_anchor_list(search_anchor_list, new_text)
        search_option_list = re.findall(r'\{\%.+?\%\}', new_text)  # Все описанное выше только для якоря с опцией
        new_text = MyOdtCheckerAnchor._check_anchor_list(search_option_list, new_text)
        return new_text




from .my_odt_checker_anchor import MyOdtCheckerAnchor
from .setting_extend_diagram import SettingExtendDiagram
from .setting_extend_set import SettingExtendSet
from .setting_add_images import SettingAddImages
from .odt_apply_settings import ODTApplySettings
from .setting_if_block import SettingIfBlock



from typing import NamedTuple, Optional
from ..config import ConfigurationMyOdt
from . import SettingExtendSet
from . import SettingAddImages
from . import SettingExtendDiagram
from . import SettingIfBlock
from ..lib import ODTMeta

# logger = ConfigurationMyOdt.logger


class ODTApplySettings:
    """A class that aggregates all document settings. Applies the necessary parameters and settings."""
    def __init__(self, file_name: str, temp_path: str, date: ODTMeta):
        self._file_name = file_name
        self._temp_path = temp_path
        self._date = date

    def loop_settings(self) -> None:
        """Function for applying settings to a template"""
        # Применяем настройки к шаблону
        try:
            if self._date.if_options:
                content_path = self._temp_path + self._file_name + '/content.xml'
                with open(content_path) as file_in:
                    text = file_in.read()
                if_settings = SettingIfBlock(text, self._date)
                text = if_settings.apply_setting()
                with open(content_path, 'w') as file_out:
                    file_out.write(text)
            if self._date.extend_set_text:
                for select_option in self._date.options:
                    if 'extend_set' in select_option:  # определяем есть ли опция extend_set в полученных данных
                        content_path = self._temp_path + self._file_name + '/content.xml'
                        with open(content_path) as file_in:
                            text = file_in.read()
                        setting_extend_set = SettingExtendSet(text, self._date)
                        text = setting_extend_set.apply_setting()
                        with open(content_path, 'w') as file_out:
                            file_out.write(text)
            if self._date.diagram:  # определяем есть ли опция diagram в полученных данных
                path = self._temp_path + self._file_name
                diagram = SettingExtendDiagram(path, self._date)
                diagram.apply_setting()
            if self._date.images:
                content_path = self._temp_path + self._file_name + '/content.xml'
                path_to_folder = self._temp_path + self._file_name
                with open(content_path) as file_in:
                    text = file_in.read()
                setting_add_images = SettingAddImages(text, self._date, path_to_folder)
                setting_add_images.apply_setting()
        except Exception:
            # logger.exception(u'odtapplysettings > odt_apply_settings > OdtApplySettings > loop_settings > Error')
            raise



# -*- coding: utf-8 -*-
import re
from .config import ConfigurationMyOdt
from .lib import Archiver, Handler, ODTMeta
from .odtapplysettings import ODTApplySettings


# logger = ConfigurationMyOdt.logger


class MyOdfGenerator:
    """
    "ENG"
    Handler for *.ODT format templates. Allows you to create templates and replace anchors of the "{{STR}}
    "type with the passed values in "data_list". The "data_list" sheet must match the pattern: data_list = [{{'key_':
    'STR', 'render_text': 'You're text ...] where" key_ "contains the anchor name from the template, and "render_text"
    the value to replace.

    "RUS"
    Обработчик шаблонов формата *.ODT. Позваляет создавать
    шаблоны, и заменять в них якоря типа "{{STR}}" на переданные значения в "data_list". Лист "data_list" должен
    соответствовать образцу: data_list = [{{'key_': 'STR', 'render_text': 'You're text ...] где "key_" содержит имя
    якоря из шаблона, а  "render_text" значение для замены.tg'
    """
    def __init__(self, file_name_without_file_extension: str, path_to_file: str, out_path: str, meta: ODTMeta):
        self._file_name = Handler.check_filename(file_name_without_file_extension)
        self._path_to_file = Handler.check_slash(path_to_file)
        self._temp_path = self._path_to_file + 'temp/'
        self._out_path = Handler.check_slash(out_path)
        self._data = meta

    def move_and_unpack(self) -> None:
        """The method moves the template file to a temporary directory and unpacks it."""
        file_move = self._path_to_file + self._file_name + '.odt'
        Handler.movefile(file_move, self._temp_path, self._file_name, '.zip')
        Archiver.extract_template(self._temp_path, self._file_name, self._temp_path)
        Handler.removefile(self._temp_path + self._file_name + '.zip')

    def settings_apply(self) -> None:
        try:
            setting = ODTApplySettings(self._file_name, self._temp_path, self._data)
            setting.loop_settings()
        except Exception:
            # logger.exception('odt_engine > MyOdfGenerator > settings_apply > File not render')
            raise

    def _clear_template(self, text: str) -> str:
        """Method for processing the template and replacing key markers with the passed values."""
        # Ищем переменные и заменяем их значения
        try:
            list_clear = re.findall(r'\{\{.+?\}\}', text)
            if len(list_clear) > 0:
                for i in list_clear:
                    text = text.replace(i, '')
            return text
        except Exception:
            # logger.exception('odt_engine > MyOdfGenerator > _clear_template > File not clear')
            raise

    def render_template(self) -> None:
        """Method for processing the template and replacing key markers with the passed values."""
        # Ищем переменные и заменяем их значения
        content_path = self._temp_path + self._file_name + '/content.xml'
        try:
            with open(content_path) as file_in:
                text = file_in.read()
            # Меняем ключь значение по тексту.
            for i in self._data.text_replace:
                if 'key_' and 'render_text' in i:
                    text = text.replace('{{' + i['key_'] + '}}', i['render_text'])
            # Бежим по опции extend_set, перебираем все строки.
            if self._data.extend_set_text:
                for content_extend_set in self._data.extend_set_text:
                    for row_content in content_extend_set['content']:
                        for i in row_content:
                            if 'key_' and 'render_text' in i:
                                text = text.replace('{{' + i['key_'] + '}}', i['render_text'])
            text = self._clear_template(text)
            with open(content_path, 'w') as file_out:
                file_out.write(text)
        except Exception:
            # logger.exception('odt_engine > MyOdfGenerator > _render_template > File not render')
            raise

    def move_and_pack_template(self) -> str:
        """The method packages the template files and moves them to the output folder."""
        # Запаковывем результат в архив .zip
        try:
            path_to_pack = self._temp_path + self._file_name
            Archiver.pack_template(path_to_pack, self._out_path, self._file_name)
            Handler.removefile(path_to_pack)
            old_name = self._out_path + self._file_name + '.zip'
            new_name = self._out_path + self._file_name + '.odt'
            Handler.renamefile(old_name, new_name)
            return new_name
        except Exception:
            # logger.exception(u'odt_engine > MyOdfGenerator > _move_and_pack_template > File zip not complete')
            raise










from typing import Optional
from .lib import Handler, ODTMeta, runtime
from .config import ConfigurationMyOdt
from .odt_engine import MyOdfGenerator
import re
import datetime

# logger = ConfigurationMyOdt.logger


class ODTLoader:
    """The class loader data. Prepares data for further work with it"""
    def __init__(self, data_dict: dict):
        self.data = data_dict  # Получаем словарь содержащий входные данные

    def _get_template_block(self) -> Optional[list]:
        """Получаем список данных которые входят шаблонный блок"""
        pass  # TODO

    def _get_if_options(self) -> Optional[dict]:
        """Получаем условия"""
        # Если опция ехтенд сет существует
        if 'if_options' in self.data['content']:
            return self.data['content']['if_options']
        else:
            return None

    def _get_extend_set_text(self) -> Optional[list]:
        """Getting a list of data to replace in tables that use the extend_set option. If there is no such data,
         None is returned"""
        # Если опция ехтенд сет существует
        if 'extend_set' in self.data['content']:
            return self.data['content']['extend_set']

        else:
            return None

    def _get_diagram(self) -> Optional[list]:
        """Getting a list of data for working with charts. If there is no such data, None is returned"""
        if 'diagram' in self.data['content']:
            return self.data['content']['diagram']
        else:
            return None

    def _get_images(self) -> Optional[list]:
        """Getting a list of data for working with images. If there is no such data, None is returned."""
        if 'images' in self.data['content']:
            return self.data['content']['images']
        else:
            return None

    def _get_extend_set(self) -> Optional[dict]:
        """We get a list of options and their settings. If there is no such data, None is returned.
        [{'extend_set': {'count': 2, 'len1': 3, 'len2': 6}}]"""
        extend_set = {}
        if 'extend_set' in self.data['content']:
            count = int(len(self.data['content']['extend_set']))
            extend_set['count'] = count
            for i in range(count):
                len_str = len(self.data['content']['extend_set'][i]['content'])
                extend_set[self.data['content']['extend_set'][i]['key_']] = len_str
            return extend_set
        else:
            return None

    def _get_opt_diagram(self) -> Optional[dict]:
        if 'diagram' in self.data['content']:
        #  'diagram': {'diagram1': {'row': 3, 'cell': 4}, 'diagram2': {'row': 3, 'cell': 4}}
            temp_diagram_dict = {}
            for diagram in self.data['content']['diagram']:
                row = len(diagram['content'])
                cell = len(diagram['content'][0]['cells'])
                for i in diagram['content']:
                    if cell != len(i['cells']):
                        err = 'Длины строк должны быть равны: > ' + diagram['key_']
                        # logger.exception('app > ODTLoader > _get_opt_diagram > Except: ' + err)
                        raise Exception(err)
                temp_diagram_dict[diagram['key_']] = dict(row=row, cell=cell)
            return temp_diagram_dict
        else:
            return None

    def _get_options(self) -> Optional[list]:
        temp_list = []
        temp_dict = {'extend_set': '', 'diagram': ''}
        extend_send = self._get_extend_set()
        diagram = self._get_opt_diagram()
        if extend_send:
            temp_dict['extend_set'] = extend_send
        if diagram:
            temp_dict['diagram'] = diagram
        temp_list.append(temp_dict)
        if len(temp_list) > 0:
            return temp_list
        else:
            temp_list = None
            return temp_list

    def load(self) -> ODTMeta:
        """Base method of the class. Use it for work!"""
        try:
            name = self.data['document_name']
            document = self.data['document_template_binary']
            text_replace = self.data['content']['text_and_table_content']
            extend_set_text = self._get_extend_set_text()
            diagram = self._get_diagram()
            images = self._get_images()
            options = self._get_options()
            if_options = self._get_if_options()
            meta = ODTMeta(name=name, document=document, text_replace=text_replace, extend_set_text=extend_set_text,
                           diagram=diagram, images=images, options=options, if_options=if_options)
        except Exception:
            # logger.exception('app > ODTBuilder > load > Except:')
            raise
        return meta


class ODTBuilder:
    def __init__(self, meta: ODTMeta):
        self.meta_data = meta
        self.path = Handler.check_slash(ConfigurationMyOdt.path_template_folder)
        self.temp_path = self.path + 'temp/'
        self.out_path = self.path + 'out/'

    def _write_template(self) -> str:
        """Method for creating a template file from binary code"""
        Handler.check_and_make(self.path)
        new_name = self.meta_data.name
        with open(self.path + new_name + '.odt', 'wb') as file_out:
            file_out.write(self.meta_data.document)
        return new_name

    def _add_marker(self) -> None:
        """
        The method checks whether the {{REPORT_ID}} marker field is present in the template body, and
        adds it if it is not present. To later replace the marker with the request ID.
        """
        # Проверяем наличие маркера {{REPORT_ID}} в шаблоне
        search = '{{REPORT_ID}}'
        marker_var = 'xmlns:reportid="{{REPORT_ID}}"'
        content_path = self.temp_path + self.meta_data.name + '/content.xml'
        if Handler.is_exist(content_path):
            with open(content_path) as file_in:
                content = file_in.read()
            if search not in content:
                # logger.info('app > ODTBuilder > _add_marker > No marker')
                replace_text_select = re.findall(r'xmlns:.+', content)
                replace_text = marker_var + ' ' + replace_text_select[0]
                content = content.replace(replace_text_select[0], replace_text)
                with open(content_path, 'w') as file_out:
                    file_out.write(content)
        #     else:
        #         # logger.debug('app > ODTBuilder > _add_marker > Contains marker')
        # else:
        #     # logger.error(
        #     #     'app > ODTBuilder > _add_marker > {} - no such file or directory'.format(content_path))


    def build(self) -> str:
        new_name = Handler.check_filename(self._write_template())
        myodfcreate = MyOdfGenerator(new_name, self.path, self.out_path, self.meta_data)
        myodfcreate.move_and_unpack()
        self._add_marker()
        myodfcreate.settings_apply()
        myodfcreate.render_template()
        full_path_out = myodfcreate.move_and_pack_template()
        return full_path_out



class ODTemplater:
    def __init__(self, data: dict):
        self.data_dict = data
        self.full_path_out = None
        self.text_bytes = None

    def _read_template(self) -> None:
        """ Возвращает файл в виде байтовой последовательности """
        with open(self.full_path_out, 'rb') as file:
            self.text_bytes = file.read()

    @runtime
    def create(self) -> bytes:
        # logger.info('////////  ODTemplater > create > start--  ////////')
        welcome_string = '{} - ODTemplater started...'.format(datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S"))
        print(welcome_string)
        loader = ODTLoader(self.data_dict)
        load_date = loader.load()
        build_date = ODTBuilder(load_date)
        self.full_path_out = build_date.build()
        self._read_template()
        Handler.removefile(self.full_path_out)
        farewell_string = '{} - ODTemplater done.'.format(datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S"))
        print(farewell_string)
        # logger.info('////////  ODTemplater > create > end--  ////////')
        return self.text_bytes




import re
from ..config import ConfigurationMyOdt
from ..lib import Handler, ImageResize
from ..lib import ODTMeta


# logger = ConfigurationMyOdt.logger


class SettingAddImages:
    """Class for replacing template images in the processed template."""
    def __init__(self, text_xml: str, date: ODTMeta, path_to_folder_template: str):
        self.text = text_xml
        self.date = date
        self.path = Handler.check_slash(path_to_folder_template)

    def replace_img(self, name_pictures: str, bin_pictures: bytes) -> None:
        """This function overwrites the image in the document files"""
        name_master = name_pictures
        name_submissive = name_pictures.split('/')
        name_submissive[-1] = '@#$' + name_submissive[-1]
        name_submissive = '/'.join(name_submissive)
        with open(self.path + name_submissive, 'wb') as file_out:
            file_out.write(bin_pictures)
        img = ImageResize(self.path + name_master, self.path + name_submissive, self.path + name_master)
        img.apply_resize()
        Handler.removefile(self.path + name_submissive)

    def loop_search_images(self) -> None:
        """The function finds images in the document body and matches them to the template. And passes the name of
        the image in the template files for replacement. """
        img_tags = re.findall(r'<draw:frame.+?</draw:frame>', self.text)  # Ищем все изображения в тексте
        # Перебираем теги изображения
        for img_tag in img_tags:
            # Перебираем название маркеров из входных параметров.
            for images in self.date.images:
                var_anchor = '{{' + images['key_'] + '}}'
                # Если есть совпадения входных данных с текстом в документе
                if var_anchor in img_tag:
                    # Выделяем опцию содержащию имя изображения
                    name = re.findall(r'xlink:href=".+?"', img_tag)
                    # Очищаем весь мусор
                    name = name[0].replace('xlink:href="', '').replace('"', '')
                    # Вызываем функцию перезаписи изображения
                    self.replace_img(name, images['source_binary'])

    def apply_setting(self) -> None:
        try:
            self.loop_search_images()
        except Exception:
            # logger.exception(u'odtapplysettings > setting_add_images > SettingAddImages > apply_setting > Error')
            raise



from setuptools import setup, find_packages

setup(name='odtemplater',
      version='0.0.1',
      description='ODT file template engine',
      packages=['odtemplater', 'odtemplater.config', 'odtemplater.lib', 'odtemplater.odtapplysettings'],
      author_email='myzxzrus@gmail.com',
      zip_safe=False,
      install_requires=['Pillow>=8.0.1'])



import re
from ..config import ConfigurationMyOdt
from . import MyOdtCheckerAnchor
from ..lib import ODTMeta


# logger = ConfigurationMyOdt.logger


class SettingExtendSet:
    """The class uses the option to multiply rows in document tables."""
    def __init__(self, text_xml: str, date: ODTMeta):
        self.text = text_xml
        self.date = date
        self.table = []  # Объявляем пустую переменную типа list

    @staticmethod
    def _chek_index_str_sequence(list_index: list) -> bool:
        """
        "ENG" Function for checking the sequence of numbers in a sheet.
        If each subsequent number is greater than one, it returns True, but if not, it returns False.

        "RUS":" Функция для проверки последовательности чисел в листе.
        Если каждое последующее число больше на единицу, то возвращается True, если же нет то False.
        """
        # Предыдущий индекс не сущетсвует.
        prev_index = None
        # Перебираем индексы.
        for i in list_index:
            # Если предыдущий индекс не существует
            if prev_index is None:
                # Приравниваем предыдущий индекс индексу из листа.
                prev_index = i
            # Если существует
            else:
                # Выполняем проверку на числовую последовательность.
                if (i - prev_index) == 1:
                    # Если число больше предыдущего на единицу, переписываем переменную предыдущего индекса на нынешний
                    prev_index = i
                # Если же число не последовательно возвращаем False
                else:
                    return False
        # Если все числа прошли проверку возвращаем True
        return True

    def _parsing_table(self) -> list:
        """
        "ENG" The function gets a list of tables and brings it to the form: [table 1, table 2, table 3...] where
        table[i] = [start, [row 1, row 2, row 3...], end].

        "RUS" Функция получает список таблиц и приводит его к виду: [table1, table2, table3...] где
        table[i] = [start, [row1, row2, row3...], end] Начало таблицы, список строк, конец таблицы.
        """
        try:
            # Перебираем список таблиц
            for table_select in range(len(self.table)):
                """
                Находим все от начала первой до конца последней строки в таблицах. Берем индекс [0], так как
                получаем список, с одним элементом.
                """
                table_row = re.findall(r'(<table:table-row.*</table:table-row>)', self.table[table_select])[0]
                # Добавляем разделитель между строками
                table_row = table_row.replace('</table:table-row>', '</table:table-row>↑')
                #  Разбиваем стринговую переменную содержащую строки по добавленному разделителю.
                table_row = table_row.split('↑')
                #  Так как разделитель добавляется и к окончанию всех строк, то убираем последний пустой элемент списка.
                table_row = table_row[:-1]
                #  Получаем начало таблицы до <table:table-row
                table_start = re.findall(r'<table:table table:name.+?<table:table-row', self.table[table_select])[
                    0].replace('<table:table-row', '')
                #  Получаем конец таблицы от последнего закрытого тега </table:table-row>
                table_end = self.table[table_select].replace(
                    re.findall(r'.*</table:table-row>', self.table[table_select])[0], '')
                #  Делаем лист содержащий в себе начало конец таблицы и список строк такблицы.
                #  temp_list = [var1, [var1, var2, ...], var3]
                temp_list = [table_start, table_row, table_end]
                #  Заменяем в списке таблиц выбранную таблицу на таблицу со списком строк.
                self.table[table_select] = temp_list
            return self.table
        except Exception:
            # logger.exception(u'setting_extend_set > SettingExtendSet > _parsing_table > Error')
            raise

    @staticmethod
    def _apply_numeric_anchor(row_table: str, group_name: str, number_replace: str) -> str:
        """
        "ENG" The function removes service information in the anchor in the form of the group name and the ~separator.
        And changes the saliva marker # for the sequence number passed to the function. As a result, it returns a
        string with numbered and cleared anchors.

        "RUS" Функция убирает в якоре служебную информацию в виде имени группы и разделителя ~. И меняет служебюный
        маркер # на порядковый номер переданный функции. Как результат возвращает строку с пронумерованными и
        очищенными якорями."""
        # Создаем список исходных якорей
        list_anchors = re.findall(r'\{\{' + group_name + '~.+?\}\}', row_table)
        # Создаем пустой список для новых якорей
        list_new_anchors = []
        # Проверяем наличие якорей в строке.
        if len(list_anchors) > 0:
            # Перебираем якоря
            for i in list_anchors:
                # Меняем в якоре служебную информацию на необходимые значения
                # {{GR1~TEXT#}} -> {{TEXT1}}
                anchors = i.replace(group_name + '~', '').replace('#', number_replace)
                # Добавляем измененный якорь в список
                list_new_anchors.append(anchors)
            # Убеждаемся что списки равны по колличеству элементов, если True
            if len(list_anchors) == len(list_new_anchors):
                # Перебираем индексы списка один (индексы элементов равны)
                for i in range(len(list_anchors)):
                    # Меняем исходный элемент в строке на измененный.
                    row_table = row_table.replace(list_anchors[i], list_new_anchors[i])
                # Возврощаем строку
                return row_table
            # Если длины списков не равны, то возвращаем ошибку.
            else:
                raise Exception('Lists are not equal!')

    @staticmethod
    def _search_group_apply_multiplying_strings(table_string: list, group_name: str, int_multiplier: int) -> list:
        """
        "ENG" The function searches for groups of anchors and strings. Calls the anchor numbering function,
        multiplies rows, and returns new, enlarged list of strings

        "RUS" Функция ищет групы якорей и строк. Вызывает функцию нумерации якоря, размножает строки, и возвращает
        новый, увеличенный список строк
        """
        temp_string_table = []
        index_str = []
        # Перебираем полученный список строк в таблицах
        for index_row_table in range(len(table_string)):
            # Если в строке есть якорь с именем группы то добавляем ее индекс в список.
            if re.search(r'\{\{' + group_name + '~.+?\}\}', table_string[index_row_table]):
                index_str.append(index_row_table)
        # Проверим список строк на размножение на возможную ошибку в виде не последовательно стоящих строк.
        check = SettingExtendSet._chek_index_str_sequence(index_str)
        if not check:
            raise Exception('The multiplication strings are inconsistent!')
        # Перебираем индексы входного листа строк и добавляем строки во временный список.
        check_multiplier = False
        for i in range(len(table_string)):
            # Если выбранный индекс отсутсвует в списке индексов строк для размножения
            if i not in index_str:
                # Добавляем строку по индексу во временный список.
                temp_string_table.append(table_string[i])
            # Если же строка есть в списке для размножения
            else:
                # Берем последовательно каждую строчку и прибавляем ее нужное колличество раз.
                if not check_multiplier:
                    for count_multiplier in range(int_multiplier):
                        for index_row in index_str:
                            temp_str = SettingExtendSet._apply_numeric_anchor(table_string[index_row], group_name,
                                                                              str(count_multiplier + 1))
                            temp_string_table.append(temp_str)
                            check_multiplier = True
        return temp_string_table

    def _apply_setting_table(self) -> list:
        """
        "ENG" The function applies options to table rows. .

        "RUS" Функция применяет опции к строкам таблиц.
        """
        try:
            # counter_options = 0
            for i in range(len(self.table)):  # table[0] = [var1, [var1, var2, ...], var3]
                table_get = self.table[i][1]  # Берем список строк из таблиц
                for extend_gr_name in self.date.extend_set_text:
                    gr_name = extend_gr_name['key_']
                    int_multiplier = int(self.date.options[0]['extend_set'][gr_name])
                    self.table[i][1] = self._search_group_apply_multiplying_strings(table_get, gr_name, int_multiplier)
            return self.table
        except Exception:
            # logger.exception(u'odt_engine > SettingExtendSet > apply_setting > Error')
            raise

    def _build_out_xml(self) -> str:
        """
        "ENG"The function collects table sheets back into XML text

        "RUS"Функция собирает листы таблиц, обратно в текст XML
        """
        for sp_i in range(len(self.table)):
            self.table[sp_i][1] = ''.join(self.table[sp_i][1])  # объединяем строки в таблицах
            self.table[sp_i] = ''.join(self.table[sp_i])  # объединяем началок строки и конец в таблице.
        # Замещаем сатрый текст на новый в XML
        for text_iter_table in range(len(re.findall(r'(<table:table table:name.+?</table:table>)', self.text))):
            self.text = self.text.replace(
                re.findall(r'(<table:table table:name.+?</table:table>)', self.text)[text_iter_table],
                self.table[text_iter_table])
        return self.text

    def apply_setting(self) -> str:
        """
        "ENG" The function applies all the methods of the class, and returns the result.

        "RUS" Функция применяет все  методы класса, и возвращает результат.
        """
        try:
            self.text = MyOdtCheckerAnchor.check_anchor_bag(self.text)  # Исключаем ошибки форматирования якорей
            #  Находим таблицы в тексте, получаем их список.
            self.table = re.findall(r'(<table:table table:name.+?</table:table>)', self.text)
            #  Парсим таблицу
            self.table = self._parsing_table()
            #  Применяем опции
            self.table = self._apply_setting_table()
            #  Упаковываем строки обратно в текст
            self.text = self._build_out_xml()
            return self.text
        except Exception:
            # logger.exception(u'odt_engine > SettingExtendSet > apply_setting > Error')
            raise
